/**
 * Copyright (C) 2012
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 */
package org.n52.ses.startupinit;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.SocketTimeoutException;
import java.net.URL;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;

import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlObject;
import org.joda.time.DateTime;
import org.joda.time.format.ISODateTimeFormat;
import org.n52.ses.util.common.ConfigurationRegistry;
import org.n52.ses.util.common.SESProperties;


/**
 * Servlet class that wakes up the SES after a (re-)start.
 *
 */
public class StartupInitServlet extends HttpServlet {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private WakeUpThread thread;

	@Override
	public void init() throws ServletException {
		super.init();

		/*
		 * contact the SES to wake it up
		 */
		SESProperties parameters = new SESProperties();
		InputStream propStream = getServletContext().getResourceAsStream("/WEB-INF/classes/" +ConfigurationRegistry.CONFIG_FILE);
		try {
			parameters.load(propStream);
		} catch (IOException e) {
			log(e.getMessage(), e);
		}

		String sesurl = "http://localhost:8080/ses/services/Broker";
		
		try {
			sesurl = parameters.getProperty(ConfigurationRegistry.SES_INSTANCE);
		} catch (Exception e) {
			/*empty*/
		}
		
		int time = 1000;

		try {
			time = Integer.parseInt(parameters.getProperty(ConfigurationRegistry.TIME_TO_WAKEUP));
		} catch (NumberFormatException e) {
			/*empty*/
		}

		log("##Startup Init Wakeup## contacting "+ sesurl);

		this.thread = new WakeUpThread(sesurl, time);
		this.thread.start();

	}


	

	@Override
	public void destroy() {
		this.thread.setRunning(false);
		
		super.destroy();
	}

	public static String getGetCapabilitiesRequest(String sesurl) throws IOException {
		InputStream capsstream = StartupInitServlet.class.getResourceAsStream(
				"/sesconfig/wakeup_capabilities_start.xml");

		BufferedReader br = new BufferedReader(new InputStreamReader(capsstream));
		StringBuilder sb = new StringBuilder();
		while (br.ready()) {
			sb.append(br.readLine());
		}
		br.close();
		
		sb.append(sesurl);
		
		capsstream = StartupInitServlet.class.getResourceAsStream(
				"/sesconfig/wakeup_capabilities_end.xml");
		br = new BufferedReader(new InputStreamReader(capsstream));
		while (br.ready()) {
			sb.append(br.readLine());
		}
		br.close();
		
		return sb.toString();
	}


	private class WakeUpThread extends Thread {

		private String sesurl;
		private boolean running;
		private int wakeUpTime;
		private boolean firstRun = true;

		public WakeUpThread(String url, int time) {
			this.wakeUpTime = time;
			this.sesurl = url;
			this.running = true;
			this.setName("SES-WakeUp-Thread");
		}


		public void setRunning(boolean running) {
			this.running = running;
		}


		@Override
		public void run() {
			int errors = 0;
			while(this.running && errors < 10) {
				errors++;
				try {
					sendWakeUpPost();
					sendWakeUpNotification();
				} catch (IOException e) {
					log(e.getMessage(), e);
				} catch (InterruptedException e) {
					log(e.getMessage(), e);
				} catch (XmlException e) {
					log(e.getMessage(), e);
				}
			}
			
			
		}
		

		private void sendWakeUpNotification() throws MalformedURLException, IOException {
			/*
			 * send an initial wakeup notification -> main resources get initialized
			 */
			HttpURLConnection conn = (HttpURLConnection) new URL(this.sesurl).openConnection();
			conn.setUseCaches(false);
			conn.setDoOutput(true);
			conn.setConnectTimeout(1000);
			conn.setReadTimeout(1000);

			//get output writer
			OutputStreamWriter wr = new OutputStreamWriter(conn.getOutputStream());

			BufferedReader capsstream = null;
			try {
				capsstream = new BufferedReader(
						new InputStreamReader(StartupInitServlet.class.getResourceAsStream(
								"/sesconfig/wakeup_notification.xml")));
			} catch (NullPointerException e) {
				
			}
			
			if (capsstream == null) return;

			StringBuilder sb = new StringBuilder();
			while (capsstream.ready()) {
				sb.append(capsstream.readLine());
			}
			capsstream.close();
			
			//send service URL (in <wsa.To> element)
			wr.write(sb.toString().replace("${ses_host}", this.sesurl).
					replace("${now}", new DateTime().toString(ISODateTimeFormat.dateTime())));

			//flush writer
			wr.flush();
			
			if (conn.getResponseCode() != HttpURLConnection.HTTP_NO_CONTENT) {
				log("Could not send initial notification.");
			} else {
				log("Succesfully send initial notification.");
			}
			
			wr.close();			
		}


		


		private void sendWakeUpPost() throws IOException, InterruptedException, XmlException {

			if (this.firstRun) {
				Thread.sleep(this.wakeUpTime);
				this.firstRun = false;
			}
			
			//send data
			int responseCode = 500;
			String doc = "";

			int errors = 0;
			while (responseCode != 200 && errors < 5) {
				errors++;
				//open connection
				HttpURLConnection conn = (HttpURLConnection) new URL(this.sesurl).openConnection();
				conn.setUseCaches(false);
				conn.setDoOutput(true);
				conn.setDoInput(true);
				conn.setConnectTimeout(5000);
				conn.setReadTimeout(5000);
				conn.connect();

				//get output writer
				OutputStreamWriter wr = new OutputStreamWriter(conn.getOutputStream());

				//send caps
				wr.write(getGetCapabilitiesRequest(sesurl));

				//flush writer
				wr.flush();
				wr.close();
				
				try {
					responseCode = conn.getResponseCode();
				} catch (SocketTimeoutException e) {
					log("Socket timeout at host " + conn.getURL() +". Retrying... Exception was: "+ e.getMessage());
				}
				if (responseCode >= HttpURLConnection.HTTP_MULT_CHOICE) {
					Thread.sleep(1000);
					continue;
				}
				
				//get response
				BufferedReader rd = null;
				try {
					rd = new BufferedReader(new InputStreamReader(conn.getInputStream()));
				} catch (IOException e) {
					log(e.getMessage(), e);
					continue;
				}

				String line;

				while ((line = rd.readLine()) != null) {
					doc += line;
				}

				rd.close();

				if (!doc.isEmpty()) log("##response from SES:## "+ doc);
			}

			if (doc.isEmpty()) return;
			
			XmlObject xDoc = XmlObject.Factory.parse(doc);
			if (xDoc != null) {
				setRunning(false);
			}
			

		}

	}
	
}
